/****************************************************************************
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *~                             IR Kernel                                    ~
 *~                         Copyright(C) 1993                                ~
 *~                       Software Kinetics Ltd.                             ~
 *~                        All Rights Reserved                               ~
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Copyright Software Kinetics Ltd.
 * 
 * Software Kinetics Ltd. has granted a licence of this software to
 * National Library of Canada.
 *
 * The National Library of Canada grants to the user a royalty-free
 * non-exclusive licence in perpetuity to use, modify, improve,
 * reproduce and sell the software.
 *
 * The software is provided to the user "as is" without warranty of
 * any kind.  Neither Software Kinetics Ltd. nor the National Library
 * of Canada warrants,guarantees or makes any representation regarding
 * the use, or the results of the use of the software in terms of
 * accuracy,reliability, or otherwise.  The entire risk as results and
 * performance of the software is assumed by the user. Neither
 * Software Kinetics Ltd. nor the National Library of Canada shall be
 * liable for any direct, indirect, consequential or incidental
 * damages arising out the use of software.
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ 
 *
 * Program: irckdel.c
 *
 * Purpose: This file contains the functions to check the mandatory fields of 
 *          delete request PDU and delete response PDU.
 *
 * Invocation:
 *
 * Input Arguments:
 *
 * Output/Exit Values: None
 *
 * Functions used:
 *
 * Version: Date:   94/05/24
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *            IR KERNEL - INTERFACE REQUIREMENTS SPECIFICATION
 *            OSI - ANSI Z39.50 DOCUMENTATION
 *
 * Notes:
 *
 * $Id:$ 
 * $Log:$
 *
 *
 *****************************************************************************/

#include <stdio.h>
#include <rpc/rpc.h>
#include <fcntl.h>
#include <pwd.h>
#include <ctype.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include "tp.h"
#include "asn_incl.h"
#include "irerror.h"
#include "tpconst.h"
#include "tpbbox.h"

/**************************** Extern Global Ids **************************/

extern int errno;

/*****************************************************************************
 *
 * Function: ir_fill_deleteResultSet_response
 *
 * Purpose: This function fills and creates the delete response PDU.
 *
 * Input Arguments:
 *              PDU_dreq - delete request PDU pointer
 *
 * Return values: SUCCESS or FAILED
 *
 * Output Arguments:
 *              PDU_dresp - delete response PDU pointer
 *
 * External Variables:
 *
 * Function used:
 *
 * Algorithm:
 *
 * Version: Date:   94/05/24
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference:
 *
 * Notes:
 *
 ****************************************************************************/
ir_fill_deleteResultSet_response(PDU_dreq, PDU_dresp)
    APDU *PDU_dreq;
    APDU *PDU_dresp;
{

    NamePlusRecord *data_ptr;

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "Calling ir_fill_deleteResultSet_response\n");
#endif

    PDU_dresp->a.deleteResultSetResp = (struct DeleteResultSetResp *) Asn1Alloc(sizeof(struct DeleteResultSetResp));

    PDU_dresp->choiceId = APDU_DELETERESULTSETRESP;

    PDU_dresp->a.deleteResultSetResp->referenceId.octetLen = PDU_dreq->a.deleteResultSetReq->referenceId.octetLen;
    PDU_dresp->a.deleteResultSetResp->referenceId.octs = (char *) Asn1Alloc(PDU_dreq->a.deleteResultSetReq->referenceId.octetLen);
    memcpy(PDU_dresp->a.deleteResultSetResp->referenceId.octs, PDU_dreq->a.deleteResultSetReq->referenceId.octs, PDU_dreq->a.deleteResultSetReq->referenceId.octetLen);

    PDU_dresp->a.deleteResultSetResp->deleteOperStatus = 2; /* invalid mandatory field */

    PDU_dresp->a.deleteResultSetResp->deleteListStatuses = (AsnList *) Asn1Alloc(sizeof(AsnList));
    PDU_dresp->a.deleteResultSetResp->deleteListStatuses->first = NULL;
    PDU_dresp->a.deleteResultSetResp->deleteListStatuses->last = NULL;
    PDU_dresp->a.deleteResultSetResp->deleteListStatuses->curr = NULL;
    PDU_dresp->a.deleteResultSetResp->deleteListStatuses->count = 0;
    PDU_dresp->a.deleteResultSetResp->deleteListStatuses->dataSize = 0;

    PDU_dresp->a.deleteResultSetResp->numberNotDeleted = (AsnInt *) Asn1Alloc(sizeof(AsnInt));
    *PDU_dresp->a.deleteResultSetResp->numberNotDeleted = 0;

    PDU_dresp->a.deleteResultSetResp->bulkStatuses = (AsnList *) Asn1Alloc(sizeof(AsnList));
    PDU_dresp->a.deleteResultSetResp->bulkStatuses->first = NULL;
    PDU_dresp->a.deleteResultSetResp->bulkStatuses->last = NULL;
    PDU_dresp->a.deleteResultSetResp->bulkStatuses->curr = NULL;
    PDU_dresp->a.deleteResultSetResp->bulkStatuses->count = 0;
    PDU_dresp->a.deleteResultSetResp->bulkStatuses->dataSize = 0;

	PDU_dresp->a.deleteResultSetResp->deleteMessage.octetLen = sizeof("MANDATORY FIELDS ARE INCOMPLETE") - 1;
	PDU_dresp->a.deleteResultSetResp->deleteMessage.octs = (char *) Asn1Alloc(PDU_dresp->a.deleteResultSetResp->deleteMessage.octetLen);
	memcpy(PDU_dresp->a.deleteResultSetResp->deleteMessage.octs, "MANDATORY FIELDS ARE INCOMPLETE", PDU_dresp->a.deleteResultSetResp->deleteMessage.octetLen);

    return (SUCCESS);
}

/*****************************************************************************
 *
 * Function: ir_create_deleteResultSet_response
 *
 * Purpose: This function creates the delete response PDU, encodes, and
 *          creates the delete response APDU for sending to IR kernel.
 *
 * Input Arguments:
 *              PDU_dreq - delete request PDU
 *              data_ptr - temporary data buffer for encoding process
 *
 * Return values: SUCCESS or FAILED
 *
 * Output Arguments:
 *              outputBuf - encoded APDU pointer
 *              APDU_length - the size of the encoded APDU
 *
 * External Variables:
 *
 * Function used:
 *              ir_fill_deleteResultSet_response()
 *              PrintDeleteResultSetResp()
 *              SBufInstallData()
 *              SBufResetInWriteRvsMode()
 *              BEncAPDU()
 *              ir_fwrite()
 *
 * Algorithm:
 *
 * Version: Date:   94/05/24
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference:
 *
 * Notes:
 *
 ****************************************************************************/
ir_create_deleteResultSet_response(PDU_dreq, outputBuf, data_ptr, APDU_length)
	APDU *PDU_dreq;
    SBuf *outputBuf;
    char **data_ptr;
    int *APDU_length;
{
    APDU PDU_dresp;

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "Calling ir_create_deleteResultSet_response()\n");
#endif

    ir_fill_deleteResultSet_response(PDU_dreq, &PDU_dresp);

#ifdef DEBUG_LEVEL2
    fprintf(DEBUG_FILE2, "ir_create_deleteResultSet_response: begin dump data:\n");
    PrintDeleteResultSetResp(DEBUG_FILE2, PDU_dresp.a.deleteResultSetResp, 0);
    fprintf(DEBUG_FILE2, "ir_create_deleteResultSet_response: end dump data:\n");
#endif

    SBufInstallData(outputBuf, (*data_ptr), MAXMESSAGESIZE * 2);
    SBufResetInWriteRvsMode(outputBuf);

    *APDU_length = BEncAPDU(outputBuf, &PDU_dresp);

	if (SBufWriteError(outputBuf))
	{
		display_error("KERNEL", ERR_ENCODE_APDU, -1);
		return(FAILED);
	}

#ifdef DEBUG_LEVEL2
    fprintf(DEBUG_FILE2, "ir_create_deleteResultSet_response: delete response APDU Dump from tpdelete\n");
    ir_fwrite(SBufDataPtr(outputBuf), SBufDataLen(outputBuf), 1, DEBUG_FILE2);
#endif

    return (SUCCESS);
}

/*****************************************************************************
 *
 * Function: ir_actual_check_delete_request
 *
 * Purpose: This function checks the delete request PDU to make sure the
 *          mandatory fields are present and valid.
 *
 * Input Arguments: PDU - delete request PDU pointer
 *
 * Return values: SUCCESS or FAILED
 *
 * Output Arguments:
 *
 * External Variables:
 *
 * Function used:
 *              display_error()
 *
 * Algorithm:
 *
 * Version: Date:   94/05/24
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference:
 *
 * Notes:
 *
 ****************************************************************************/
ir_actual_check_delete_request(PDU_ptr)
    APDU *PDU_ptr;
{

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "Calling ir_actual_check_delete_request\n");
#endif

    if (PDU_ptr->a.deleteResultSetReq->deleteOperation != 0 &&
        PDU_ptr->a.deleteResultSetReq->deleteOperation != 1)
    {
        display_error("KERNEL", ERR_INVALID_DELETE_OPERATION, -1);
        return (FAILED);
    }

    return (SUCCESS);
}

/*****************************************************************************
 *
 * Function: ir_check_deleteResultSet_request
 *
 * Purpose: This function performs the following functions:
 *              - checks the delete request PDU,
 *              - creates and sends the delete response APDU to Origin.
 *
 * Input Arguments:
 *              PDU - delete request PDU pointer
 *              data_ptr - temporary data pointer for ASN functions
 *              socketFd - the socket Fd for communicating to IR kernel
 *
 * Return values: SUCCESS or FAILED
 *
 * Output Arguments:
 *
 * External Variables:
 *
 * Function used:
 *              ir_check_delete_request()
 *              ir_create_deleteResultSet_response()
 *              tpa_send_APDU()
 *              display_error()
 *
 * Algorithm:
 *
 * Version: Date:   94/05/24
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference:
 *
 * Notes:
 *
 ****************************************************************************/
ir_check_deleteResultSet_request(PDU_dreq, data_ptr, socketFd)
    APDU *PDU_dreq;
    char **data_ptr;
    int socketFd;
{
    int result;
    long APDU_length;
    SBuf APDU_dresp;

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "Calling ir_process_deleteResultSet_request\n");
#endif

    result = ir_actual_check_delete_request(PDU_dreq);

    if (result == FAILED)
    {
        /*
         * create a delete response with result 'mandatory fields are
         * incomplete'
         */
        ir_create_deleteResultSet_response(PDU_dreq, &APDU_dresp, data_ptr, &APDU_length);

		if (cs_send_data(APDU_dresp.dataStart, (int) APDU_length, socketFd) == FAILED)
		{
			display_error("KERNEL", ERR_SEND_DATA_TO_CLIENT, errno);
			return (FAILED);
		}
#ifdef DEBUG_LEVEL1
		else
			fprintf(DEBUG_FILE1, "ir_process_deleteResultSet_request: %ld deleteResultSet response sent\n", APDU_length);
#endif
	}

    return (SUCCESS);
}

/*****************************************************************************
 *
 * Function: ir_check_deleteResultSet_response
 *
 * Purpose: This function checks the mandatory fields of delete response PDU 
 *          for validity.
 *
 * Input Arguments: PDU_ptr - delete response PDU pointer
 *
 * Return values: SUCCESS or FAILED
 *
 * Output Arguments:
 *
 * External Variables:
 *
 * Function used:
 *              display_error()
 *
 * Algorithm:
 *
 * Version: Date:   94/05/24
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference:
 *
 * Notes:
 *
 ****************************************************************************/
ir_check_deleteResultSet_response(PDU_ptr)
    APDU *PDU_ptr;
{

    if (PDU_ptr->a.deleteResultSetResp->deleteOperStatus < 0 ||
        PDU_ptr->a.deleteResultSetResp->deleteOperStatus > 9)
    {
        display_error("KERNEL", ERR_INVALID_DELETE_OPERATION_STATUS, -1);
        return (FAILED);
	}

    return (SUCCESS);
}
