/****************************************************************************
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *~                             IR Kernel                                    ~
 *~                         Copyright(C) 1993                                ~
 *~                       Software Kinetics Ltd.                             ~
 *~                        All Rights Reserved                               ~
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Copyright Software Kinetics Ltd.
 * 
 * Software Kinetics Ltd. has granted a licence of this software to
 * National Library of Canada.
 *
 * The National Library of Canada grants to the user a royalty-free
 * non-exclusive licence in perpetuity to use, modify, improve,
 * reproduce and sell the software.
 *
 * The software is provided to the user "as is" without warranty of
 * any kind.  Neither Software Kinetics Ltd. nor the National Library
 * of Canada warrants,guarantees or makes any representation regarding
 * the use, or the results of the use of the software in terms of
 * accuracy,reliability, or otherwise.  The entire risk as results and
 * performance of the software is assumed by the user. Neither
 * Software Kinetics Ltd. nor the National Library of Canada shall be
 * liable for any direct, indirect, consequential or incidental
 * damages arising out the use of software.
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ 
 *
 * Program: irg2apdu.c
 *
 * Purpose: This file contains the function to move the second APDU to the
 *          first APDU position and to return an event.
 *
 * Invocation:
 *
 * Input Arguments:
 *
 * Output/Exit Values: None
 *
 * Functions used: ir_get_second_APDU()
 *
 * Version: Date:   93/05/12
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *            IR KERNEL - INTERFACE REQUIREMENTS SPECIFICATION
 *            OSI - ANSI Z39.50 DOCUMENTATION
 *
 * Notes:
 *
 * $Id: irg2apdu.c,v 1.2 1993/12/13 19:18:15 irsuprt Exp irsuprt $
 * $Log: irg2apdu.c,v $
 * Revision 1.2  1993/12/13  19:18:15  irsuprt
 * V1.2: added #include <error.h> for AIX_SYSTEM.
 *
 * Revision 1.1  1993/11/01  17:03:51  irsuprt
 * Initial revision
 *
 *
 *****************************************************************************/
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef AIX_SYSTEM
	#include <error.h>
#endif 
#include "ir.h"
#include "irerror.h"
#include "irassoc.h"
#include "irstatab.h"
#include "irconst.h"
#include "asn_incl.h"

/*****************************************************************************
 *
 * Function: ir_get_second_APDU
 *
 * Purpose: To move the second APDU APDU_ptr2 to the first APDU APDU_ptr
 *          position and to return an event. It resets the Id APDUsize2 to 0.
 *
 * Input Arguments: assocId - current association
 *                  PDU_ptr - decoded PDU from the 2nd APDU Id APDU_ptr2
 *
 * Return values: An valid event or zero means an error had occurred.
 *
 * Output Arguments: None
 *
 * External Variables:
 *
 * Function used:
 *                  ResetNibbleMem()
 *                  SBufInstallData()
 *                  SBufResetInReadMode()
 *                  setjmp()
 *                  BDecTag()
 *                  BDecLen()
 *                  cs_Pdata_indication()
 *                  display_error()
 *
 * Algorithm:
 *
 * Version: Date:   93/05/12
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * References:
 *
 * Notes:
 *
 ****************************************************************************/

int 
ir_get_second_APDU(assocId, PDU_ptr)
    struct assoc_struct **assocId;
    APDU *PDU_ptr;
{
    int val;
    AsnLen bsize, bsizea;
    char **buf;
    int remainingBytes;
    ENV_TYPE env;
    SBuf outputBuf;
    AsnTag tag;
    APDU apduptr;
    char *data;

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "Calling ir_get_second_APDU\n");
#endif

    /* copy the 2nd APDU to the 1st APDU position */
    memcpy((*assocId)->APDU_ptr, (*assocId)->APDU_ptr2, (*assocId)->APDUsize2);

    /* make a copy of the APDU to store in data */
    data = (char *) malloc(MAXMESSAGESIZE * 2);
    memcpy(data, (*assocId)->APDU_ptr, (*assocId)->APDUsize2);

    /* ResetNibbleMem(MAXMESSAGESIZE, MAXMESSAGESIZE); */

    /* Find out the actual size of the APDU */
    SBufInstallData(&outputBuf, data, (*assocId)->APDUsize2);
    SBufResetInReadMode(&outputBuf);

    bsizea = (AsnLen) 0;
    if ((val = setjmp(env)) == 0)
    {
        tag = BDecTag(&outputBuf, &bsizea, env);
        bsize = BDecLen(&outputBuf, &bsizea, env);
        (*assocId)->APDUsize = bsize + bsizea;
        (*assocId)->next_start_pos = (*assocId)->APDUsize2;

#ifdef DEBUG_LEVEL3
        fprintf(DEBUG_FILE3, "ir_get_second_APDU: APDU size=%ld, next_pos=%ld\n", (*assocId)->APDUsize, (*assocId)->APDUsize2);
#endif

        free(data);

        if ((*assocId)->APDUsize > (*assocId)->APDUsize2)
        {
            /* There is more data coming for this APDU */
            (*assocId)->APDUsize2 = 0;
            return (0);
        }
        else
        {
            (*assocId)->APDUsize2 = 0;
            /* The APDU is complete, decode it. */
            if (cs_Pdata_indication((*assocId)->APDU_ptr, PDU_ptr, (*assocId)->APDUsize) == SUCCESS)
                return (ir_map_event(PDU_ptr));
            else
                return (IR_ABORT_REQUEST);
        }
    }
    else
    {
        display_error("KERNEL", ERR_FIND_APDU_LENGTH, errno);
        (*assocId)->APDUsize2 = 0;
        free(data);
        return (0);
    }

}                                /* end of procedure ir_get_second_APDU */
