/****************************************************************************
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *~                             IR Kernel                                    ~
 *~                         Copyright(C) 1993                                ~
 *~                       Software Kinetics Ltd.                             ~
 *~                        All Rights Reserved                               ~
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Copyright Software Kinetics Ltd.
 * 
 * Software Kinetics Ltd. has granted a licence of this software to
 * National Library of Canada.
 *
 * The National Library of Canada grants to the user a royalty-free
 * non-exclusive licence in perpetuity to use, modify, improve,
 * reproduce and sell the software.
 *
 * The software is provided to the user "as is" without warranty of
 * any kind.  Neither Software Kinetics Ltd. nor the National Library
 * of Canada warrants,guarantees or makes any representation regarding
 * the use, or the results of the use of the software in terms of
 * accuracy,reliability, or otherwise.  The entire risk as results and
 * performance of the software is assumed by the user. Neither
 * Software Kinetics Ltd. nor the National Library of Canada shall be
 * liable for any direct, indirect, consequential or incidental
 * damages arising out the use of software.
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ 
 *
 * Program: tp_init.c
 *
 * Purpose: This file contains the functions to create the init target process,
 *          create init response and send it to IR kernel and then fork/exec
 *          QRM process. This is being used with QRM not the black box.
 *
 * Invocation:
 *
 * Input Arguments:
 *
 * Output/Exit Values: None
 *
 * Functions used:
 *                  tpa_ALI_setup()
 *                  tp_wait_for_connect()
 *                  tp_fill_init_response()
 *                  tp_create_init_response()
 *                  tpa_send_APDU()
 *                  open_files()
 *                  close_files()
 *                  main()
 *                  create_QRM()
 *
 * Version: Date:   93/05/12
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *            IR KERNEL - INTERFACE REQUIREMENTS SPECIFICATION
 *            OSI - ANSI Z39.50 DOCUMENTATION
 *
 * Notes:
 *
 * $Id: tp_init.c,v 1.1 1993/11/01 17:03:51 irsuprt Exp $
 * $Log: tp_init.c,v $
 * Revision 1.1  1993/11/01  17:03:51  irsuprt
 * Initial revision
 *
 * Revision 1.1  1993/11/01  17:03:51  irsuprt
 * Initial revision
 *
 *
 *****************************************************************************/
#include <stdio.h>
#include <rpc/rpc.h>
#include <fcntl.h>
#include <pwd.h>
#include <ctype.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/socket.h>

#ifndef VMS
#include <netinet/in.h>
#endif

#include <netdb.h>
#include <arpa/inet.h>
#include "ir.h"
#include "asn_incl.h"
#include "irerror.c"
#include "irconst.h"

#define PERMS 0666
#define MAXLEN          50
#define FAILED           -1
#define MaxQueueSize      5

char IMPLEMENTATION_ID[] = "IR Z39.50 Kernel V1.0";
char IMPLEMENTATION_NAME[] = "National Library IR Z39.50 Kernel";
char IMPLEMENTATION_VERSION[] = "1.0";

FILE *DEBUG_FILE1, *DEBUG_FILE2, *DEBUG_FILE3;

extern int errno;

/*****************************************************************************
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *~                             SR Kernel                                    ~
 *~                         Copyright (C) 1992                               ~
 *~                       Software Kinetics Ltd.                             ~
 *~                        All Rights Reserved                               ~
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * Function:    tpa_ALI_setup
 *
 * Purpose:     This module sets up the TCP IR Service port.
 *
 * Input Args:
 *
 * Return
 *     Values:  None.
 *
 *      $Log: tp_init.c,v $
 * Revision 1.1  1993/11/01  17:03:51  irsuprt
 * Initial revision
 *
 * Revision 1.1  1993/11/01  17:03:51  irsuprt
 * Initial revision
 *
 * Revision 1.1  1993/11/01  16:07:34  irsuprt
 * Initial revision
 *
 * Revision 1.1  1993/11/01  16:04:37  irsuprt
 * Initial revision
 *
 * Revision 0.9  92/11/18  12:06:00  saunders
 *
 *****************************************************************************/

int
tpa_ALI_setup(tpPort, socketFd)
    int *tpPort;
    int *socketFd;
{
    struct sockaddr_in serverAddress, clientAddress;
    struct hostent *hostInfo;
    char hostname[MAXLEN];
    int length;

    /* Open a TCP Socket to communicate with a client */
    *socketFd = socket(AF_INET, SOCK_STREAM, 0);

    if (*socketFd < 0)
    {
        display_error("TP_INIT", ERR_CREATE_SOCKET, -1);
        return (FAILED);
    }

    /* Bind our local address so that a client can send to us */
    bzero((char *) &serverAddress, sizeof(serverAddress));
    serverAddress.sin_family = AF_INET;

    /* Get our hostname */
    if (gethostname(hostname, MAXLEN) < 0)
    {
        display_error("TP_INIT", ERR_CANNOT_GET_HOST_NAME, errno);
        return (FAILED);
    }
    hostInfo = gethostbyname(hostname);

    if (hostInfo != 0)
    {

#ifdef VMS
        serverAddress.sin_addr.s_addr = INADDR_ANY;
#else
        serverAddress.sin_addr.s_addr = inet_network(inet_ntoa(*hostInfo->h_addr_list));
#endif
    }
    else
    {
        display_error("TP_INIT", ERR_CANNOT_GET_HOST_NAME, errno);
        return (FAILED);
    }
    serverAddress.sin_port = 0;

    length = sizeof(serverAddress);

    if (bind(*socketFd, (struct sockaddr *) & serverAddress, length) < 0)
    {
        display_error("TP_INIT", ERR_BIND_SOCKET, -1);
        return (FAILED);
    }

    if (getsockname(*socketFd, (struct sockaddr *) & serverAddress, &length) < 0)
        display_error("TP_INIT", ERR_GET_SOCKET_NAME, -1);

    *tpPort = serverAddress.sin_port;

    listen(*socketFd, MaxQueueSize);

    return (SUCCESS);

}                                /* end of tpa_ALI_setup  */

tp_wait_for_connect(socketFd)
    int *socketFd;
{
    struct sockaddr_in serverAddress;
    int length, new_socketFd;

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "tp_wait_for_connect: Waiting for Connection .....\n");
#endif

    /* Wait for a connection from a Client process */
    length = sizeof(serverAddress);
    new_socketFd = accept(*socketFd, (struct sockaddr *) & serverAddress, &length);
    if (new_socketFd < 0)
    {
        display_error("TP_INIT", ERR_ACCEPT_CONNECT, -1);
        return (FAILED);
    }
    else
    {
        *socketFd = new_socketFd;

#ifdef DEBUG_LEVEL3
        fprintf(DEBUG_FILE3, "tp_wait_for_connect: accepted connect %d\n", *socketFd);
#endif
    }
    return (SUCCESS);
}

tp_fill_init_response(PDU_ptr, socketFd)
    APDU *PDU_ptr;
    int socketFd;
{

    PDU_ptr->a.initResp = (struct InitializeResp *) malloc(sizeof(struct InitializeResp));

    PDU_ptr->choiceId = APDU_INITRESP;

    PDU_ptr->a.initResp->referenceId.octs = (char *) malloc(6);
    sprintf(PDU_ptr->a.initResp->referenceId.octs, "%5d", socketFd);
    PDU_ptr->a.initResp->referenceId.octetLen = strlen(PDU_ptr->a.initResp->referenceId.octs) + 1;

    PDU_ptr->a.initResp->protocolVersion.bitLen = 2;
    PDU_ptr->a.initResp->protocolVersion.bits = malloc(2);
    strcpy(PDU_ptr->a.initResp->protocolVersion.bits, "1");

    PDU_ptr->a.initResp->options.bitLen = 8;
    PDU_ptr->a.initResp->options.bits = malloc(8);
    strcpy(PDU_ptr->a.initResp->options.bits, "1111111");

    PDU_ptr->a.initResp->preferredMessageSize = MAXMESSAGESIZE;
    PDU_ptr->a.initResp->maximumRecordSize = MAXMESSAGESIZE;
    PDU_ptr->a.initResp->result = TRUE;        /* accept */

    PDU_ptr->a.initResp->implementationId.octs = malloc(sizeof(IMPLEMENTATION_ID) + 1);
    strcpy(PDU_ptr->a.initResp->implementationId.octs, IMPLEMENTATION_ID);
    PDU_ptr->a.initResp->implementationId.octetLen = strlen(PDU_ptr->a.initResp->implementationId.octs) + 1;

    PDU_ptr->a.initResp->implementationName.octs = malloc(sizeof(IMPLEMENTATION_NAME) + 1);
    strcpy(PDU_ptr->a.initResp->implementationName.octs, IMPLEMENTATION_NAME);
    PDU_ptr->a.initResp->implementationName.octetLen = strlen(PDU_ptr->a.initResp->implementationName.octs) + 1;

    PDU_ptr->a.initResp->implementationVersion.octs = malloc(sizeof(IMPLEMENTATION_VERSION) + 1);
    strcpy(PDU_ptr->a.initResp->implementationVersion.octs, IMPLEMENTATION_VERSION);
    PDU_ptr->a.initResp->implementationVersion.octetLen = strlen(PDU_ptr->a.initResp->implementationVersion.octs) + 1;

}

tp_create_init_response(socketFd, outputBuf, data_ptr, APDU_length)
    int socketFd;
    SBuf *outputBuf;
    char **data_ptr;
    int *APDU_length;
{
    APDU PDU_ptr;
    int i;

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "Calling tp_create_init_response()\n");
#endif

    tp_fill_init_response(&PDU_ptr, socketFd);

    /*
     * SBufInit(outputBuf, (*data_ptr), MAXMESSAGESIZE*2);
     * SBufResetInWriteRvsMode(outputBuf);
     */

    InitNibbleMem(MAXMESSAGESIZE, MAXMESSAGESIZE);

    SBufInstallData(outputBuf, (*data_ptr), MAXMESSAGESIZE * 2);
    SBufResetInWriteRvsMode(outputBuf);

    *APDU_length = BEncAPDU(outputBuf, &PDU_ptr);

#ifdef DEBUG_LEVEL2

#ifndef VMS
    fprintf(DEBUG_FILE2, "Init Response APDU Dump from tp_init\n");
    ir_fwrite(SBufDataPtr(outputBuf), SBufDataLen(outputBuf), 1, DEBUG_FILE2);
#endif

#endif
}

int
tpa_send_APDU(socketFd, data_ptr, APDU_length)
    int socketFd;
    SBuf *data_ptr;
    int APDU_length;
{

    int i, numSend;

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "Calling tpa_send_APDU(), APDU_length=%d\n", APDU_length);
#endif

    /*
     * for (i=0; i<APDU_length; i++) printf("%c=%x,",
     * *(data_ptr->dataStart+i), *(data_ptr->dataStart+i)); printf("\n");
     */

    numSend = send(socketFd, data_ptr->dataStart, APDU_length, 0);

#ifdef DEBUG_LEVEL3
    fprintf(DEBUG_FILE3, "tpa_send_APDU: %d bytes sent\n", numSend);
#endif

    return (numSend);

}

open_files()
{
    char tty_no[BUFSIZ];
    FILE *f;
    int failure = 0;

    if (!(f = fopen("ttyno", "r")))
    {
        perror("TP_INIT - error opening `ttyno', using stdout");
        failure = 1;
    }
    else
    {
        fscanf(f, "%s", tty_no);
        fclose(f);
    }

    printf("TP_INIT - open_file: tty_no=%s\n", tty_no);

#if (defined(DEBUG_LEVEL2) || defined(DEBUG_LEVEL3)) && !defined(DEBUG_LEVEL1)
#error "DEBUG_LEVEL1 must be defined if DEBUG_LEVEL2 or DEBUG_LEVEL3 is defined"
#endif

#ifdef DEBUG_LEVEL1
    if (failure || ((DEBUG_FILE1 = fopen(tty_no, "a")) == NULL))
        DEBUG_FILE1 = stdout;
#endif

#ifdef DEBUG_LEVEL2

#ifdef VMS
    DEBUG_FILE2 = DEBUG_FILE1;
#else
    if ((DEBUG_FILE2 = fopen(tty_no, "a")) == NULL)
        return (FAILED);
#endif

#endif

#ifdef DEBUG_LEVEL3

#ifdef VMS
    DEBUG_FILE3 = DEBUG_FILE1;
#else
    if ((DEBUG_FILE3 = fopen(tty_no, "a")) == NULL)
        return (FAILED);
#endif

#endif

    return (SUCCESS);
}                                /* End Of Procedure ir_open_files */

close_files()
{
    if (DEBUG_FILE1)
        fclose(DEBUG_FILE1);

#ifndef VMS
    if (DEBUG_FILE2)
        fclose(DEBUG_FILE2);
    if (DEBUG_FILE3)
        fclose(DEBUG_FILE3);
#endif
}

main(argc, argv)
    int argc;
    char *argv[];
{
    FILE *writefd;
    char *pipename, buff[255];
    int tpPort;
    int socketFd;
    int res;
    SBuf APDU_ptr;
    int APDU_length;
    char *data_ptr;


    if (argc > 0)
    {

        if (tpa_ALI_setup(&tpPort, &socketFd) != SUCCESS)
            tpPort = 0;

#ifdef VMS
        printf("TP_INIT: tpPort is %d, about to open pipe...\n", tpPort);
        /* Get file descriptor # for pipe from the command-line */
        if ((writefd = fdopen(atoi(argv[1]), "w")) == NULL)
#else
        pipename = argv[1];        /* Create named pipe */
        if ((writefd = fopen(pipename, "w")) == NULL)
#endif

        {
            display_error("TP_INIT", ERR_OPEN_WRITE_FILE, -1);
            exit(1);
        }
        else
        {
            fprintf(writefd, "%d\n", tpPort);
            fclose(writefd);
        }

        /* ` */
        open_files();

#ifdef DEBUG_LEVEL1
        fprintf(DEBUG_FILE1, "Running tp_init\n");
#endif

        if (tp_wait_for_connect(&socketFd) == SUCCESS)
        {
            /*
             * res = recv(socketFd, buff, sizeof(buff), 0);
             * printf("\ntp_init: main: res=%d, buff=%s\n", res, buff);
             */

            /*
             * strcpy(buff, "Hello"); res = send(socketFd, buff,
             * sizeof(buff), 0); printf("\ntp_init: main sending message
             * Hello res=%d\n", res);
             * 
             * printf("tp_init: main: before calling malloc\n");
             */

            data_ptr = malloc(MAXMESSAGESIZE * 2);

#ifdef DEBUG_LEVEL1
            fprintf(DEBUG_FILE1, "tp_init: main: before calling tp_create_init_response\n");
#endif

            tp_create_init_response(socketFd, &APDU_ptr, &data_ptr, &APDU_length);
            if (tpa_send_APDU(socketFd, &APDU_ptr, APDU_length) <= 0)
            {
                display_error("TP_INIT", ERR_SEND_DATA_TO_IR_FROM_TP, errno);
            }
            else
            {
                close(socketFd);
                create_QRM(socketFd);
            }
        }

        netclose(socketFd);
        close_files();

    }                                /* end if */

#ifdef VMS
    /* Wait for any child processes to terminate before we exit. */
    /* If we don't, we'll take them down with us when we exit.   */
    while (wait(NULL) != -1)
        ;
#endif

    exit(0);
}

#ifndef VMS
create_QRM(socketFd)
    int socketFd;
{
    char pipename[16];
    int pid, new_socketFd;
    FILE *writefd;

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "tp_init: create_QRM\n");
#endif

    pid = fork();
    if (pid == 0)
    {                                /* child process */
        printf("execl qrmhandler");
        execl("qrmhandler", "qrmhandler", NULL, 0);        /* start qrmhandler */
    }                                /* end if child */

    return;

    /*******************************************************************/
    sprintf(pipename, "/tmp/pipef%05d", getpid());
    printf("Pipe Name: %s\n", pipename);
    if ((mknod(pipename, S_IFIFO | PERMS | O_NDELAY, 0) < 0) && (errno != EEXIST))
    {
        printf("Could not create FIFO.\n");
    }
    else
    {
        pid = fork();
        if (pid == 0)
        {                        /* child process */
            printf("execl tp_QRM\n");
            execl("tp_QRM", "tp_QRM", pipename, 0);        /* start qrmhandler */
        }                        /* end if child */

        if ((writefd = fopen(pipename, "w")) == NULL)
            printf("Could not open FIFO for writing.\n");
        new_socketFd = dup(socketFd);
        fprintf(writefd, "%d\n", new_socketFd);
        printf("create_QRM: new_socketFd=%d\n", new_socketFd);
        fclose(writefd);
        unlink(pipename);

    }
}
#endif
